#   Copyright ETH 2023 Zürich, Scientific IT Services
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import concurrent.futures

import pandas as pd

from pybis.property_reformatter import is_of_openbis_supported_date_format
from .openbis_command import OpenbisCommand
from ..command_result import CommandResult
from ..utils import cd
from ...scripts.click_util import click_echo


def _dfs(objects, prop, func, func_specific):
    """Helper function that perform DFS search over children graph of objects"""
    # TODO: improve performance of this - make it similar to _dfs_samples
    with concurrent.futures.ThreadPoolExecutor(
            max_workers=5) as pool_simple, concurrent.futures.ThreadPoolExecutor(
        max_workers=20) as pool_full:
        stack = [openbis_obj[prop] for openbis_obj in
                 objects]  # datasets and samples provide children in different formats
        visited = set()
        stack.reverse()
        output = []
        while stack:
            simple_results = pool_simple.map(func, stack)
            stack = []
            children = []
            full_download = []
            for obj in simple_results:
                key = obj.df[prop][0]
                children += list(obj.df['children'])[0]
                if key not in visited:
                    visited.add(key)
                    full_download += [key]
            if full_download:
                output += pool_full.map(func_specific, full_download)
            for child in children:
                if child not in visited:
                    stack += [child]

    return output


def _dfs_samples(data_base, prop, func):
    """Helper function that perform DFS search over children graph of objects"""
    output = data_base
    ids = [x['children'] for x in data_base if x['children']]
    ids = [x[prop][prop] for x in flatten(ids)]
    visited = set([x[prop][prop] for x in data_base])
    while ids:
        data = func(ids)
        data = list(data.values())
        output += data
        ids = []
        children = []
        for obj in data:
            key = obj[prop][prop]
            children += [x[prop][prop] for x in obj['children']]
            if key not in visited:
                visited.add(key)
        for child in children:
            if child not in visited:
                ids += [child]
    return output


def flatten(matrix):
    flat_list = []
    for row in matrix:
        flat_list += row
    return flat_list


def _check_date(sign, date1, date2):
    if is_of_openbis_supported_date_format(date1) and is_of_openbis_supported_date_format(date2):
        timestamp1 = pd.to_datetime(date1)
        timestamp2 = pd.to_datetime(date2)
        if sign == "=":
            return timestamp2 == timestamp1
        elif sign == ">":
            return timestamp2 > timestamp1
        elif sign == "<":
            return timestamp2 < timestamp1
        raise ValueError(f"Unknown sign {sign}")
    else:
        raise ValueError("Dates are not in a supported formats!")


def _filter_dataset(dataset, filters):
    if filters.get("space", None) is not None:
        space = filters["space"]
        if dataset.sample is not None and dataset.sample.space.code != space:
            return False
        if dataset.experiment is not None and dataset.experiment.project.space.code != space:
            return False
    if filters.get("type_code", None) is not None:
        if dataset.type.code != filters["type_code"]:
            return False
    if filters.get("project", None) is not None:
        project = filters["project"]
        if dataset.sample is not None and dataset.sample.project.code != project:
            return False
        if dataset.experiment is not None and dataset.experiment.project.code != project:
            return False
    if filters.get("experiment", None) is not None:
        if dataset.experiment is not None and dataset.experiment.code != filters["experiment"]:
            return False
    if filters.get("property_code", None) is not None:
        prop_code = filters["property_code"]
        prop_value = filters["property_value"]
        if dataset.props is not None and dataset.props[prop_code.lower()] != prop_value:
            return False
    if filters.get("registration_date", None) is not None:
        registration_date = filters["registration_date"]
        sign = "="
        if registration_date[0] in [">", "<", "="]:
            sign, registration_date = registration_date[0], registration_date[1:]
        return _check_date(sign, registration_date, dataset.registrationDate)
    if filters.get("modification_date", None) is not None:
        modification_date = filters["modification_date"]
        sign = "="
        if modification_date[0] in [">", "<", "="]:
            sign, modification_date = modification_date[0], modification_date[1:]
        return _check_date(sign, modification_date, dataset.modificationDate)
    return True


class Search(OpenbisCommand):
    """
    Command to search samples or datasets in openBIS.
    """

    def __init__(self, dm, filters, recursive, save_path):
        """
        :param dm: data management
        :param filters: Dictionary of filter to be used during search
        :param recursive: Flag indicating recursive search in children
        :param save_path: Path to save results. If not set, results will not be saved.
        """
        self.filters = filters
        self.recursive = recursive
        self.save_path = save_path
        self.load_global_config(dm)
        self.props = "*"
        super(Search, self).__init__(dm)

    def search_samples(self):
        search_results = self._search_samples(raw_response=True)

        search_results = self.openbis._sample_list_for_response(props=self.props,
                                                                response=search_results,
                                                                attrs=["parents", "children",
                                                                       "dataSets"],
                                                                parsed=True)

        click_echo(f"Objects found: {len(search_results)}")
        if self.save_path is not None:
            click_echo(f"Saving search results in {self.save_path}")
            with cd(self.data_mgmt.invocation_path):
                search_results.df.to_csv(self.save_path, index=False)
        else:
            click_echo(f"Search results:\n{search_results}")

        return CommandResult(returncode=0, output="Search completed.")

    def _get_samples_children(self, identifier):
        return self.openbis.get_samples(identifier, attrs=["children", "dataSets"])

    def _get_sample_with_datasets(self, identifier):
        return self.openbis.get_sample(identifier, withDataSetIds=True, raw_response=True,
                                       props=self.props)

    def _search_samples(self, raw_response=False):
        """Helper method to search samples"""

        if self.recursive:
            raw_response = True

        if "object_code" in self.filters:
            results = self.openbis.get_samples(identifier=self.filters['object_code'],
                                               attrs=["parents", "children", "dataSets"],
                                               raw_response=raw_response,
                                               props=self.props)
        else:
            args = self._get_filtering_args(self.props, ["parents", "children", "dataSets"])
            args["raw_response"] = raw_response
            results = self.openbis.get_samples(**args)

        if self.recursive:
            click_echo("Recursive search enabled. It may take time to produce results.")
            output2 = _dfs_samples(results['objects'], 'identifier', self._get_sample_with_datasets)

            search_results = output2
        else:
            if raw_response:
                search_results = results['objects']
            else:
                search_results = results
        return search_results

    def _get_datasets_children(self, permId):
        return self.openbis.get_datasets(permId, attrs=["children"])

    def search_data_sets(self):
        if self.save_path is not None and self.fileservice_url() is None:
            return CommandResult(returncode=-1,
                                 output="Configuration fileservice_url needs to be set for download.")

        main_filters = self.filters.copy()

        object_filters = {k[7:]: v for (k, v) in main_filters.items() if k.startswith('object_')}
        dataset_filters = {k: v for (k, v) in main_filters.items() if not k.startswith('object_')}
        if object_filters:
            if 'id' in object_filters:
                if object_filters['id'] is not None:
                    object_filters['object_code'] = object_filters['id']
                del object_filters['id']
            self.filters = object_filters
            search_results = self._search_samples(raw_response=True)
            click_echo(f"Samples found: {len(search_results)}")

            datasets = [x["dataSets"] for x in search_results]
            datasets = flatten(datasets)
            datasets = [x['permId']['permId'] for x in datasets]
            datasets = self.openbis.get_dataset(permIds=datasets, props=self.props)

            filtered_datasets = []
            for dataset in datasets:
                if _filter_dataset(dataset, dataset_filters):
                    filtered_datasets += [dataset]

            datasets = self.openbis._dataset_list_for_response(props=self.props,
                                                               response=[x.data for x in
                                                                         filtered_datasets],
                                                               parsed=True)
        else:
            if self.recursive:
                search_results = self._search_samples()  # Look for samples recursively
                o = []
                for sample in search_results.objects:  # get datasets
                    o += sample.get_datasets(
                        attrs=["parents", "children"], props=self.props)
                output = _dfs(o, 'permId',  # datasets provide permIds as children
                              self._get_datasets_children,
                              self.openbis.get_dataset)  # look for child datasets
                datasets = self.openbis._dataset_list_for_response(props=self.props,
                                                                   response=[dataset.data for
                                                                             dataset
                                                                             in output],
                                                                   parsed=True)
            else:
                if "dataset_id" in self.filters:
                    results = self.openbis.get_sample(self.filters['dataset_id']).get_datasets(
                        attrs=["parents", "children"], props=self.props)
                else:
                    args = self._get_filtering_args(self.props, ["parents", "children"])
                    results = self.openbis.get_datasets(**args)
                datasets = results

        click_echo(f"Data sets found: {len(datasets)}")
        if self.save_path is not None:
            click_echo(f"Saving search results in {self.save_path}")
            with cd(self.data_mgmt.invocation_path):
                datasets.df.to_csv(self.save_path, index=False)
        else:
            click_echo(f"Search results:\n{datasets}")

        return CommandResult(returncode=0, output="Search completed.")

    def _get_filtering_args(self, props, attrs):
        where = None
        if self.filters['property_code'] is not None and self.filters['property_value'] is not None:
            where = {
                self.filters['property_code']: self.filters['property_value'],
            }

        args = dict(space=self.filters['space'],
                    project=self.filters['project'],
                    # Not Supported with Project Samples disabled
                    experiment=self.filters['collection'],
                    type=self.filters['type_code'],
                    where=where,
                    attrs=attrs,
                    props=props)

        if self.filters['registration_date'] is not None:
            args['registrationDate'] = self.filters['registration_date']
        if self.filters['modification_date'] is not None:
            args['modificationDate'] = self.filters['modification_date']
        return args



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   Copyright ETH 2018 - 2023 Zürich, Scientific IT Services
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
#   
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import getpass
from unittest.mock import Mock, MagicMock, ANY

from .openbis_command import OpenbisCommand
from .. import data_mgmt


def test_prepare_run(monkeypatch):
    # given
    dm = data_mgmt.DataMgmt(openbis=Mock(), git_config={
        'data_path': '',
        'metadata_path': '',
        'invocation_path': ''
    })
    openbis_command = OpenbisCommand(dm)
    define_config(openbis_command)
    monkeypatch.setattr(getpass, 'getpass', lambda s: 'password')
    dm.openbis.is_session_active.return_value = False
    # when
    openbis_command.prepare_run()
    # then
    dm.openbis.is_session_active.assert_called()
    dm.openbis.login.assert_called_with('watney', 'password', save_token=True)


def define_config(openbis_command):
    openbis_command.config_dict = {
        'config': {
            'user': 'watney'
        }
    }
    



#   Copyright ETH 2018 - 2023 Zürich, Scientific IT Services
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
#   
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import os

from .openbis_command import OpenbisCommand, ContentCopySelector
from ..checksum import validate_checksum
from ..command_result import CommandResult, CommandException
from ..repository_utils import copy_repository
from ..utils import cd
from ... import dm


class Clone(OpenbisCommand):
    """
    Implements the clone command. Copies a repository from a content copy of a data set
    and adds the local copy as a new content copy using the addref command.
    Validates the checksums of the copied files unless skip_integrity_check is True.
    """

    def __init__(self, dm, data_set_id, ssh_user, content_copy_index, skip_integrity_check):
        """
        :param dm: data management
        :param data_set_id: permId of the data set to be cloned
        :param ssh_user: user for remote access
        :param content_copy_index: in case of multiple content copied
        :param skip_integrity_check: Checksums are not validated if True
        """
        if dm.data_path != dm.metadata_path:
            raise CommandException(CommandResult(returncode=-1,
                                                 output='Clone/move not supported with obis_metadata_folder.'))
        self.data_set_id = data_set_id
        self.ssh_user = ssh_user
        self.content_copy_index = content_copy_index
        self.skip_integrity_check = skip_integrity_check
        self.load_global_config(dm)
        super(Clone, self).__init__(dm)

    def check_configuration(self):
        missing_config_settings = []
        if self.openbis is None:
            missing_config_settings.append('openbis_url')
        if self.user() is None:
            missing_config_settings.append('user')
        if len(missing_config_settings) > 0:
            return CommandResult(returncode=-1,
                                 output="Missing configuration settings for {}.".format(
                                     missing_config_settings))
        return CommandResult(returncode=0, output="")

    def run(self):

        result = self.prepare_run()
        if result.failure():
            return result

        data_set = self.openbis.get_dataset(self.data_set_id)

        content_copy = ContentCopySelector(data_set, self.content_copy_index).select()
        self.content_copy = content_copy
        host = content_copy['externalDms']['address'].split(':')[0]
        path = content_copy['path']
        repository_folder = path.split('/')[-1]

        result = copy_repository(self.ssh_user, host, path)
        if result.failure():
            return result
        result = self.checkout_commit(content_copy, path)
        if result.failure():
            return result
        data_set = self.openbis.get_dataset(self.data_set_id)
        if self.skip_integrity_check != True:
            data_path = os.path.join(self.data_mgmt.data_path, repository_folder)
            invalid_files = validate_checksum(self.openbis, data_set.file_list, data_set.permId,
                                              data_path, self.data_mgmt.metadata_path)
            if len(invalid_files) > 0:
                raise CommandException(CommandResult(returncode=-1,
                                                     output="Invalid checksum for files {}.".format(
                                                         str(invalid_files))))
        return self.add_content_copy_to_openbis(repository_folder)

    def checkout_commit(self, content_copy, path):
        """
        Checks out the commit of the content copy from which the clone was made 
        in case there are newer commits / data sets. So the new copy is based on the 
        data set given as an input.
        """
        commit_hash = content_copy['gitCommitHash']
        repository_folder = path.split('/')[-1]
        return self.git_wrapper.git_checkout(commit_hash, relative_repo_path=repository_folder)

    def add_content_copy_to_openbis(self, repository_folder):
        with cd(repository_folder):
            data_path = os.path.join(self.data_mgmt.data_path, repository_folder)
            metadata_path = os.path.join(self.data_mgmt.metadata_path, repository_folder)
            invocation_path = self.data_mgmt.invocation_path
            data_mgmt = dm.DataMgmt(openbis_config={}, git_config={
                'find_git': True,
                'data_path': data_path,
                'metadata_path': metadata_path,
                'invocation_path': invocation_path
            })
            data_mgmt.set_property(data_mgmt.debug, data_mgmt.settings_resolver.config, 'hostname',
                                   None, False)
            return data_mgmt.addref()



#   Copyright ETH 2018 - 2023 Zürich, Scientific IT Services
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
#   
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import os
import shutil

from .checksum import ChecksumGeneratorCrc32, ChecksumGeneratorGitAnnex
from .utils import run_shell
from ..scripts.click_util import click_echo


class GitWrapper(object):
    """A wrapper on commands to git and git annex."""

    def __init__(self, git_path=None, git_annex_path=None, find_git=None, data_path=None,
                 metadata_path=None, invocation_path=None):
        self.git_path = git_path
        self.git_annex_path = git_annex_path
        self.data_path = data_path
        self.metadata_path = metadata_path

    def _git(self, params, strip_leading_whitespace=True, relative_repo_path=''):
        """ all git invocations need to go through this method
        since it sets --work-tree and '--git-dir.
         """
        cmd = [self.git_path]
        if self.data_path is not None and self.metadata_path is not None:
            git_dir = os.path.join(self.metadata_path, relative_repo_path, '.git')
            cmd += ['--work-tree', self.data_path, '--git-dir', git_dir]
        cmd += params
        return run_shell(cmd, strip_leading_whitespace=strip_leading_whitespace)

    def can_run(self):
        """Return true if the perquisites are satisfied to run (git and git annex)"""
        if self.git_path is None:
            click_echo('No git path found!')
            return False
        if self.git_annex_path is None:
            click_echo('No git-annex path found!')
            return False
        if self._git(['help']).failure():
            click_echo('Can not run git!')
            # git help should have a returncode of 0
            return False
        if self._git(['annex', 'help']).failure():
            click_echo('Can not run git-annex!')
            # git help should have a returncode of 0
            return False
        result = run_shell([self.git_path, 'annex', 'version'])
        if result.success():
            first_line = result.output.split("\n")[0].split(":")
            if len(first_line) > 1:
                self.annex_version = first_line[1].strip()
                try:
                    self.annex_major_version = int(self.annex_version.split(".")[0])
                except Exception as e:
                    print("Invalid git-annex version line:", result.output)
                    return False
        return True

    def git_init(self):
        result = self._git(["init"])
        self.git_ignore('.obis')
        self.git_ignore('.obis_restorepoint')
        return result

    def git_status(self, path=None):
        if path is None:
            return self._git(["annex", "status"], strip_leading_whitespace=False)
        else:
            return self._git(["annex", "status", path], strip_leading_whitespace=False)

    def git_annex_init(self, desc, git_annex_backend=None):
        """ Configures annex in a git repository."""

        # We use annex --version=5 since that works better with big files. Version 
        # 6 can lead to out of memory errors.
        cmd = ["annex", "init", "--version=5"]
        if desc is not None:
            cmd.append(desc)
        result = self._git(cmd)
        if result.failure():
            return result

        # annex.thin to avoid copying big files
        cmd = ["config", "annex.thin", "true"]
        result = self._git(cmd)
        if result.failure():
            return result

        # direct mode so annex uses hard links instead of soft links
        cmd = ["annex", "unlock" if self.annex_major_version >= 8 else "direct"]
        result = self._git(cmd)
        if result.failure():
            return result

        # re-enable the repository to be used with git directly
        # though we need to know what we do since annex can lead to unexpected behaviour
        cmd = ["config", "--unset", "core.bare"]
        result = self._git(cmd)
        if result.failure():
            return result

        # copy out annex config and change the annex backend according to obis config
        attributes_src = os.path.join(os.path.dirname(__file__), "git-annex-attributes")
        attributes_dst = '.git/info/attributes'
        shutil.copyfile(attributes_src, attributes_dst)
        self._apply_git_annex_backend(attributes_dst, git_annex_backend)

        return result

    def initial_commit(self):
        # initial commit is needed. we can restore to it when something fails
        return self._git(['commit', '--allow-empty', '-m', 'Initial commit.'])

    def _apply_git_annex_backend(self, filename, git_annex_backend):
        if git_annex_backend is not None:
            lines = []
            file = open(filename, "r")
            for line in file:
                if "annex.backend" in line:
                    lines.append("* annex.backend=" + git_annex_backend + "\n")
                else:
                    lines.append(line)
            file.close()
            file = open(filename, "w")
            file.writelines(lines)
            file.close()

    def git_add(self, path):
        # git annex add to avoid out of memory error when adding files bigger than RAM
        cmd = ["annex", "add", path]
        if self.annex_major_version < 8:
            cmd.append("--include-dotfiles")
        return self._git(cmd)

    def git_commit(self, msg):
        return self._git(['commit', '--allow-empty', '-m', msg])

    def git_top_level_path(self):
        return self._git(['rev-parse', '--show-toplevel'])

    def git_commit_hash(self):
        return self._git(['rev-parse', '--short', 'HEAD'])

    def git_ls_tree(self):
        return self._git(['ls-tree', '--full-tree', '-r', 'HEAD'])

    def git_checkout(self, path_or_hash, relative_repo_path=''):
        if relative_repo_path:
            return self._git(['checkout', path_or_hash], relative_repo_path=relative_repo_path)
        return self._git(["checkout", path_or_hash])

    def git_reset_to(self, commit_hash):
        return self._git(['reset', commit_hash])

    def git_ignore(self, path):
        result = self._git(['check-ignore', path])
        if result.returncode == 1:
            with open(".git/info/exclude", "a") as gitignore:
                gitignore.write(path)
                gitignore.write("\n")


class GitRepoFileInfo(object):
    """Class that gathers checksums and file lengths for all files in the repo."""

    def __init__(self, git_wrapper):
        self.git_wrapper = git_wrapper

    def contents(self, git_annex_hash_as_checksum=False):
        """Return a list of dicts describing the contents of the repo.
        :return: A list of dictionaries
          {'crc32': checksum,
           'checksum': checksum other than crc32
           'checksumType': type of checksum
           'fileLength': size of the file,
           'path': path relative to repo root.
           'directory': False
          }"""
        files = self.file_list()
        cksum = self.cksum(files, git_annex_hash_as_checksum)
        return cksum

    def file_list(self):
        tree = self.git_wrapper.git_ls_tree()
        if tree.failure() or len(tree.output) == 0:
            return []
        lines = tree.output.split("\n")
        files = [line.split("\t")[-1].strip() for line in lines]
        return files

    def cksum(self, files, git_annex_hash_as_checksum=False):

        if git_annex_hash_as_checksum == False:
            checksum_generator = ChecksumGeneratorCrc32(self.git_wrapper.data_path,
                                                        self.git_wrapper.metadata_path)
        else:
            checksum_generator = ChecksumGeneratorGitAnnex(self.git_wrapper.data_path,
                                                           self.git_wrapper.metadata_path)

        checksums = []

        for file in files:
            checksum = checksum_generator.get_checksum(file)
            checksums.append(checksum)

        return checksums



#   Copyright ETH 2018 - 2023 Zürich, Scientific IT Services
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
#   
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import os
import sys

if sys.version_info < (3, 3):
    sys.exit("Sorry, Python < 3.3 is not supported")

from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

# include all man files
data_dir = os.path.join("man", "man1")
data_files = [
    (d, [os.path.join(d, f) for f in files]) for d, folders, files in os.walk(data_dir)
]

setup(
    name="obis",
    version="0.4.4rc2",
    description="Local data management with assistance from OpenBIS.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://sissource.ethz.ch/sispub/openbis/tree/master/app-openbis-command-line",
    author="ID SIS • ETH Zürich",
    author_email="openbis-support@id.ethz.ch",
    license="Apache Software License Version 2.0",
    packages=["obis", "obis.dm", "obis.dm.commands", "obis.scripts"],
    data_files=data_files,
    package_data={"obis": ["dm/git-annex-attributes"]},
    install_requires=["pyOpenSSL", "pytest", "pybis==1.36.2rc2", "click"],
    entry_points={"console_scripts": ["obis=obis.scripts.cli:main"]},
    zip_safe=False,
    python_requires=">=3.3",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)



#   Copyright ETH 2018 - 2023 Zürich, Scientific IT Services
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
#   
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
class CommandResult(object):
    """Encapsulate result from a subprocess call."""

    def __init__(self, completed_process=None, returncode=None, output=None, strip_leading_whitespace=True):
        """Convert a completed_process object into a ShellResult."""
        if completed_process:
            self.returncode = completed_process.returncode
            if completed_process.stderr:
                self.output = completed_process.stderr.decode('utf-8').rstrip()
            else:
                self.output = completed_process.stdout.decode('utf-8').rstrip()
            if strip_leading_whitespace:
                self.output = self.output.strip()
        else:
            self.returncode = returncode
            self.output = output

    def __str__(self):
        return "CommandResult({},{})".format(self.returncode, self.output)

    def __repr__(self):
        return "CommandResult({},{})".format(self.returncode, self.output)

    def success(self):
        return self.returncode == 0

    def failure(self):
        return not self.success()


class CommandException(Exception):
    """ Instead of returning a CommandResult for all actions, we can also return 
    a result normally and throw a CommandException instead on error. """
    
    def __init__(self, command_result):
        self.command_result = command_result



#   Copyright ETH 2018 - 2023 Zürich, Scientific IT Services
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
#   
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import os
import subprocess
from contextlib import contextmanager
from datetime import datetime
from enum import Enum

from .command_result import CommandResult, CommandException


class Type(Enum):
    """Enum representing which what type of repository it is"""
    LINK = 1,
    PHYSICAL = 2,
    UNKNOWN = 3


class OperationType(Enum):
    """Enum representing basic operation types"""
    GET = 1,
    SET = 2,
    CLEAR = 3


def complete_openbis_config(config, resolver, local_only=True):
    """Add default values for empty entries in the config."""
    config_dict = resolver.config.config_dict(local_only)
    if config.get('url') is None:
        config['url'] = config_dict['openbis_url']
    if config.get('verify_certificates') is None:
        config['verify_certificates'] = config_dict['verify_certificates']
    if config.get('token') is None:
        config['token'] = None
    if config.get('is_physical') is None and config_dict['is_physical'] is not None:
        config['is_physical'] = config_dict['is_physical']
    if config.get(
            'allow_http_but_do_not_use_this_in_production_and_only_within_safe_networks') is None:
        config['allow_http_but_do_not_use_this_in_production_and_only_within_safe_networks'] = not \
            config_dict['allow_only_https']


def complete_git_config(config):
    """Add default values for empty entries in the config."""

    find_git = config['find_git'] if config.get('find_git') is not None else True
    if find_git:
        git_cmd = locate_command('git')
        if git_cmd.success():
            config['git_path'] = git_cmd.output

        git_annex_cmd = locate_command('git-annex')
        if git_annex_cmd.success():
            config['git_annex_path'] = git_annex_cmd.output


def default_echo(details):
    if details.get('level') != "DEBUG":
        print(details['message'])


def run_shell(args, shell=False, strip_leading_whitespace=True, raise_exception_on_failure=False):
    result = CommandResult(
        subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=shell),
        strip_leading_whitespace=strip_leading_whitespace)
    if raise_exception_on_failure == True and result.failure():
        raise CommandException(result)
    return result


def locate_command(command):
    """Return a tuple of (returncode, stdout)."""
    # Need to call this command in shell mode so we have the system PATH
    result = run_shell(['type {}'.format(command)], shell=True)
    # 'type -p' not supported by all shells, so we do it manually
    if result.success():
        result.output = result.output.split(" ")[-1]
    return result


@contextmanager
def cd(newdir):
    """Safe cd -- return to original dir after execution, even if an exception is raised."""
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)


def is_valid_perm_id(name):
    if "-" not in name:
        return False
    split = name.split("-")
    try:
        datetime.strptime(split[0], "%Y%m%d%H%M%S%f")
        int(split[1])
        return True
    except ValueError:
        return False
